@@ -42,8 +42,6 @@
 
 /* Logging configuration for a function */
 struct _fn_config {
-    Char* name;
-
     Int dump_before;
     Int dump_after;
     Int zero_before;
@@ -58,8 +56,6 @@ struct _fn_config {
 #if CT_ENABLE_DEBUG
     Int verbosity; /* Change debug verbosity level while in function */
 #endif
-
-    fn_config* next;
 };
 
 /* Configurations for function name prefix patterns.
@@ -82,25 +78,25 @@ struct _fn_config {
 /* node of compressed trie search structure */
 typedef struct _config_node config_node;
 struct _config_node {
-    Char* name;
-    Int length;
+  Int length;
     
-    fn_config* config;
-    config_node* sub_node[NODE_DEGREE];
-    config_node* cpp_sep;
+  fn_config* config;
+  config_node* sub_node[NODE_DEGREE];
+  config_node* next;
+  config_node* wild_star;
+  config_node* wild_char;
+
+  Char name[1];
 };
 
 /* root of trie */
 static config_node* fn_configs = 0;
 
 static __inline__ 
-fn_config* new_fnc(Char* name)
+fn_config* new_fnc()
 {
    fn_config* new = VG_(malloc)(sizeof(fn_config));
 
-   new->name = VG_(strdup)(name);
-   new->next = 0;
-
    new->dump_before  = CONFIG_DEFAULT;
    new->dump_after   = CONFIG_DEFAULT;
    new->zero_before  = CONFIG_DEFAULT;
@@ -114,170 +110,201 @@ fn_config* new_fnc(Char* name)
    new->verbosity    = CONFIG_DEFAULT;
 #endif
 
-   CT_DEBUG(2, "  new_fnc('%s')\n", name);
-
    return new;
 }
 
 
-static fn_config* get_fnc1(fn_config** pfnc, Char* name)
-{
-    fn_config* fnc = *pfnc;
-    while(fnc) {
-	if (VG_(strcmp)(fnc->name,name) == 0) break;
-	fnc = fnc->next;
-    }
-    if (!fnc) {
-	fnc = new_fnc(name);
-	fnc->next = *pfnc;
-	*pfnc = fnc;
-    }
-    return fnc;
-}
-
-
 static config_node* new_config(Char* name, int length)
 {
     int i;
-    config_node* node = VG_(malloc)(sizeof(config_node));
+    config_node* node = VG_(malloc)(sizeof(config_node) + length);
+
+    for(i=0;i<length;i++) {
+      if (name[i] == 0) break;
+      node->name[i] = name[i];
+    }
+    node->name[i] = 0;
 
-    node->name   = name;
     node->length = length;
     node->config = 0;
     for(i=0;i<NODE_DEGREE;i++)
 	node->sub_node[i] = 0;
-    node->cpp_sep = 0;
+    node->next = 0;
+    node->wild_char = 0;
+    node->wild_star = 0;
 
-    CT_DEBUG(2, "  new_config('%s', len %d)\n", name, length);
+    CT_DEBUG(3, "   new_config('%s', len %d)\n", node->name, length);
 
     return node;
 }
 
-/* Search first wildcard, with optionally appended cpp separator ("::").
- * if not found, *(return) is 0
- */
-static Char* next_wild_cpp_sep(Char* n)
+static __inline__
+Bool is_wild(Char n)
 {
-    while(*n!=0) {
-	if (*n == '*') {
-	    n++;
-	    if (*n == ':') {
-		n++;
-		if (*n == ':') {
-		    n++;
-		}
-	    }
-	    return n;
-	}
-	n++;
-    }
-    return n;
+  return (n == '*') || (n == '?');
 }
 
-/* Get function config for a specific name in search tree */
-static fn_config* get_fnc2(config_node** pnode, Char* name, int offset)
+/* Recursively build up function matching tree (prefix tree).
+ * Returns function config object for pattern <name>
+ * and starting at tree node <*pnode>.
+ *
+ * Tree nodes (config_node) are created as needed,
+ * tree root is stored into <*pnode>, and the created
+ * leaf (fn_config) for the given pattern is returned.
+ */
+static fn_config* get_fnc2(config_node* node, Char* name)
 {
-    config_node* node;
-    fn_config* fnc;
-    Char* cpp_sep;
+  config_node *new_sub, *n, *nprev;
+  int offset, len;
 
-    CT_ASSERT(pnode != 0);
-    node = *pnode;
+  CT_DEBUG(3, "  get_fnc2(%p, '%s')\n", node, name);
 
-    if (node == 0) {
-	fnc = new_fnc(name);
-	cpp_sep = next_wild_cpp_sep(name+offset);
-	if (*cpp_sep) {
-	    /* found wildcard */
-	    int newOffset = cpp_sep - name;
-	    node = new_config(name, offset);
-	    node->config = fnc;
-	    *pnode = node;
-	    CT_DEBUG(3, "  get_config('%s', off %d, new %d): cpp_sep\n",
-		     name, offset, newOffset);
-	    /* recurse */
-	    return get_fnc2( &(node->cpp_sep), cpp_sep, newOffset);
-	}
-	node = new_config(fnc->name, VG_(strlen)(fnc->name));
-	node->config = fnc;
-	*pnode = node;
+  if (name[0] == 0) {
+    if (!node->config) node->config = new_fnc();
+    return node->config;
+  }
 
-	CT_DEBUG(3, "  get_config('%s', offset %d): new\n",
-		 name, offset);
+  if (is_wild(*name)) {
+    if (*name == '*') {
+      while(name[1] == '*') name++;
+      new_sub = node->wild_star;
+    }
+    else
+      new_sub = node->wild_char;
 
-	return fnc;
+    if (!new_sub) {
+      new_sub = new_config(name, 1);
+      if (*name == '*')
+	node->wild_star = new_sub;
+      else
+	node->wild_char = new_sub;
+    }
+
+    return get_fnc2( new_sub, name+1);
+  }
+
+  n = node->sub_node[ name[0]%NODE_DEGREE ];
+  nprev = 0;
+  len = 0;
+  while(n) {
+    for(len=0; name[len] == n->name[len]; len++);
+    if (len>0) break;
+    nprev = n;
+    n = n->next;
+  }
+
+  if (!n) {
+    len = 1;
+    while(name[len] && (!is_wild(name[len]))) len++;
+    new_sub = new_config(name, len);
+    new_sub->next = node->sub_node[ name[0]%NODE_DEGREE ];
+    node->sub_node[ name[0]%NODE_DEGREE ] = new_sub;	
+
+    if (name[len] == 0) {
+      new_sub->config = new_fnc();
+      return new_sub->config;
     }
     
-    if (name[offset] == 0) {
-	CT_DEBUG(3, "  get_config('%s', off %d): found\n", name, offset);
-	return get_fnc1(&node->config, name);
+    /* recurse on wildcard */
+    return get_fnc2( new_sub, name+len);
+  }
+
+  if (len < n->length) {
+
+    /* split up the subnode <n> */
+    config_node *new_node;
+    int i;
+
+    new_node = new_config(n->name, len);
+    if (nprev)
+      nprev->next = new_node;
+    else
+      node->sub_node[ n->name[0]%NODE_DEGREE ] = new_node;
+    new_node->next = n->next;
+
+    new_node->sub_node[ n->name[len]%NODE_DEGREE ] = n;
+
+    for(i=0, offset=len; offset < n->length; i++, offset++)
+      n->name[i] = n->name[offset];
+    n->name[i] = 0;
+    n->length = i;
+
+    name += len;
+    offset = 0;
+    while(name[offset] && (!is_wild(name[offset]))) offset++;
+    new_sub  = new_config(name, offset);
+    /* this sub_node of new_node could already be set: chain! */
+    new_sub->next = new_node->sub_node[ name[0]%NODE_DEGREE ];
+    new_node->sub_node[ name[0]%NODE_DEGREE ] = new_sub;
+
+    if (name[offset]==0) {
+      new_sub->config = new_fnc();
+      return new_sub->config;
     }
 
-    if (name[offset] == '*') {
-	/* the "::" is optional in the pattern */
-	offset++;
-	if (name[offset] == ':') {
-	    offset++;
-	    if (name[offset] == ':') offset++;
-	}
-	CT_DEBUG(3, "  get_config('%s', off %d): cpp_sep\n", name, offset);
-	return get_fnc2( &(node->cpp_sep), name+offset, 0);
-    }
+    /* recurse on wildcard */
+    return get_fnc2( new_sub, name+offset);
+  }
 
-    /* FIXME: Should handle wildcard */
-    while(offset < node->length) {
-	if ((name[offset] == 0)  ||
-	    ((name[offset]%NODE_DEGREE) != (node->name[offset]%NODE_DEGREE)))
-	    break;
-	offset++;
-    }
+  name += n->length;
 
-    /* name is shorter or different as nodes name */
-    if (offset < node->length) {
-	config_node* new;
+  if (name[0] == 0) {
+    /* name and node name are the same */
+    if (!n->config) n->config = new_fnc();
+    return n->config;
+  }
 
-	/* split up this node */
-	fnc = new_fnc(name);
-	new = new_config(node->name, offset);
+  offset = 1;
+  while(name[offset] && (!is_wild(name[offset]))) offset++;
 
-	new->sub_node[ node->name[offset]%NODE_DEGREE ] = node;
-	    
-	if (name[offset]==0) {
-	    /* no subnode, insert into new node */
-	    new->config = fnc;
-	}
-	else {
-	    config_node* new2 = new_config(node->name, VG_(strlen)(name));
-	    new->sub_node[ name[offset]%NODE_DEGREE ] = new2;
-	    new2->config = fnc;
-	}
-	*pnode = new;
+  new_sub = new_config(name, offset);
+  new_sub->next = n->sub_node[ name[offset]%NODE_DEGREE ];
+  n->sub_node[ name[offset]%NODE_DEGREE ] = new_sub;	
 
-	CT_DEBUG(3, "  get_config('%s', off %d): splitted\n", name, offset);
-
-	return fnc;
-    }
-
-    /* name and node name are congruent */
-    if (name[offset] == 0) {
-	CT_DEBUG(3, "  get_config('%s', off %d): found\n", name, offset);
-	return get_fnc1(&node->config, name);
-    }
-
-    /* name is longer than the nodes name: append new node */
-    CT_DEBUG(3, "  get_config('%s', off %d): next\n", name, offset);
-
-    /* recurse */
-    return get_fnc2( &(node->sub_node[ name[offset]%NODE_DEGREE ]),
-		     name, offset);
+  return get_fnc2(new_sub, name+offset);
 }
 
-/* get a function config for a given prefix name */
+static void print_config_node(int s, config_node* node)
+{
+  config_node* n;
+  int i;
+
+  if (node != fn_configs) {
+    char sp[] = "                                        ";
+
+    if (s>40) s=40;
+    VG_(printf)(sp+40-s);
+    VG_(printf)("'%s'/%d\n", node->name, node->length);
+  }
+  for(i=0;i<NODE_DEGREE;i++) {
+    n = node->sub_node[i];
+    while(n) {
+      print_config_node(s+1, n);
+      n = n->next;
+    }
+  }
+  if (node->wild_char) print_config_node(s+1, node->wild_char);
+  if (node->wild_star) print_config_node(s+1, node->wild_star);
+}
+
+/* get a function config for a name pattern (from command line) */
 static fn_config* get_fnc(Char* name)
 {
-    return get_fnc2(&fn_configs, name, 0);
+  fn_config* fnc;
+
+  CT_DEBUG(3, " +get_fnc(%s)\n", name);
+  if (fn_configs == 0)
+    fn_configs = new_config(name, 0);
+  fnc =  get_fnc2(fn_configs, name);
+
+  CT_DEBUGIF(3) {
+    CT_DEBUG(3, " -get_fnc(%s):\n", name);
+    print_config_node(3, fn_configs);
+  }
+  return fnc;
 }
 
+  
 
 static void update_fn_config1(fn_node* fn, fn_config* fnc)
 {
@@ -311,61 +338,38 @@ static void update_fn_config1(fn_node* fn, fn_config* fnc)
 #endif
 }
 
-/* Search first cpp separator ("::").
- * if not found, m[return] is 0
+/* Recursively go down the function matching tree,
+ * looking for a match to <name>. For every matching leaf,
+ * <fn> is updated with the pattern config.
  */
-static Int next_cpp_sep(Char* n)
-{
-    Int p = 0;
-    while(n[p]!=0) {
-	if (n[p] == ':') {
-	    p++;
-	    if (n[p] == ':') {
-		p++;
-		return p;
-	    }
-	}
-	p++;
-    }
-    return p;
-}
-
 static void update_fn_config2(fn_node* fn, char* name, config_node* node)
 {
-    fn_config* fnc;
-    int len, cpp_sep_pos = -1;
+    config_node* n;
 
-    len = next_cpp_sep(name);
-    if (name[len]) {
-	cpp_sep_pos = len;
-	while(name[len]) len++;
-
-	CT_DEBUG(3, "  update_fn_config('%s', off %d): cpp_sep\n",
-		 name, cpp_sep_pos);
+    CT_DEBUG(3, "  update_fn_config2('%s', node '%s'): \n",
+	     name, node->name);
+    if ((*name == 0) && node->config) {
+      CT_DEBUG(3, "Found!\n");
+      update_fn_config1(fn, node->config);
+      return;
     }
 
-    while(node && (node->length <= len)) {
-	fnc = node->config;
-	while(fnc) {
-	    if (VG_(strncmp)(fnc->name,name,node->length) == 0) break;
-	    fnc = fnc->next;
-	}
+    n = node->sub_node[ name[0]%NODE_DEGREE ];
+    while(n) {
+      if (VG_(strncmp)(name, n->name, n->length)==0) break;
+      n = n->next;
+    }
+    if (n) update_fn_config2(fn, name+n->length, n);
     
-	if (fnc) {
-	    CT_DEBUG(3, "  update_fn_config('%s', len %d): found\n",
-		     fnc->name, node->length);
+    if (node->wild_char)
+      update_fn_config2(fn, name+1, node->wild_char);
 
-	    update_fn_config1(fn, fnc);
-
-	    /* recurse on wildcard patterns (depth first search).
-	     * Better would be to build a state machine...
-	     */
-	    if (node->cpp_sep && cpp_sep_pos>0)
-		update_fn_config2(fn, name + cpp_sep_pos, node->cpp_sep);
-	}
-
-	if (name[node->length] == 0) break;
-	node = node->sub_node[ name[node->length]%NODE_DEGREE ];
+    if (node->wild_star) {
+      while(*name) {
+	update_fn_config2(fn, name, node->wild_star);
+	name++;
+      }
+      update_fn_config2(fn, name, node->wild_star);
     }
 }
 
@@ -373,7 +377,8 @@ static void update_fn_config2(fn_node* fn, char* name, config_node* node)
 void SK_(update_fn_config)(fn_node* fn)
 {
     CT_DEBUG(3, "  update_fn_config('%s')\n", fn->name);
-    update_fn_config2(fn, fn->name, fn_configs);
+    if (fn_configs)
+      update_fn_config2(fn, fn->name, fn_configs);
 }
 
 
@@ -409,10 +414,10 @@ Bool SK_(process_cmd_line_option)(Char* arg)
    else if (0 == VG_(strcmp)(arg, "--trace-jump=no"))
        SK_(clo).collect_jumps = False;
 
-   else if (0 == VG_(strcmp)(arg, "--separate-dumps=yes"))
-       SK_(clo).separate_dumps = True;
-   else if (0 == VG_(strcmp)(arg, "--separate-dumps=no"))
-       SK_(clo).separate_dumps = False;
+   else if (0 == VG_(strcmp)(arg, "--combine-dumps=yes"))
+       SK_(clo).combine_dumps = True;
+   else if (0 == VG_(strcmp)(arg, "--combine-dumps=no"))
+       SK_(clo).combine_dumps = False;
 
    else if (0 == VG_(strcmp)(arg, "--collect-data=yes"))
        SK_(clo).collect_data = True;
@@ -568,6 +573,14 @@ Bool SK_(process_cmd_line_option)(Char* arg)
    else if (0 == VG_(strcmp)(arg, "--dump-bb=no"))
        SK_(clo).dump_bb = False;
 
+   else if (0 == VG_(strcmp)(arg, "--dump-objs=yes")) {
+       SK_(clo).dump_objs = True;
+       SK_(clo).dump_line = True;
+       SK_(clo).dump_instr = True;
+   }
+   else if (0 == VG_(strcmp)(arg, "--dump-objs=no"))
+       SK_(clo).dump_objs = False;
+
    else if (0 == VG_(strncmp)(arg, "--dump-every-bb=", 16))
         SK_(clo).dump_every_bb = (Int)VG_(atoll)(&arg[16]);
 
@@ -597,13 +610,13 @@ void SK_(print_usage)(void)
 "    --dump-instr=no|yes       Dump instruction address of costs? [no]\n"
 "    --compress-strings=no|yes Compress strings in profile dump? [yes]\n"
 "    --compress-pos=no|yes     Compress positions in profile dump? [yes]\n"
+"    --combine-dumps=no|yes    Concat all dumps into same file [no]\n"
 #if CT_EXPERIMENTAL
 "    --compress-events=no|yes  Compress events in profile dump? [no]\n"
 "    --dump-bb=no|yes          Dump basic block address of costs? [no]\n"
 "    --dump-bbs=no|yes         Dump basic block info? [no]\n"
 "    --dump-skipped=no|yes     Dump info on skipped functions in calls? [no]\n"
 "    --mangle-names=no|yes     Mangle separation into names? [yes]\n"
-"    --separate-dumps=no|yes   Write each dump into separate file [yes]\n"
 #endif
 
 "\n   activity options (for interactivity use callgrind_control):\n"
@@ -611,6 +624,7 @@ void SK_(print_usage)(void)
 "    --dump-before=<func>      Dump when entering function\n"
 "    --zero-before=<func>      Zero all costs when entering function\n"
 "    --dump-after=<func>       Dump when leaving function\n"
+"    --dump-objs=no|yes        Dump static object information [no]\n"
 
 "\n   data collection options:\n"
 "    --instr-atstart=no|yes    Do instrumentation at callgrind start [yes]\n"
@@ -666,7 +680,7 @@ void SK_(set_clo_defaults)(void)
 
   /* dump options */
   SK_(clo).filename_base    = 0;
-  SK_(clo).separate_dumps   = True;
+  SK_(clo).combine_dumps    = False;
   SK_(clo).compress_strings = True;
   SK_(clo).compress_mangled = False;
   SK_(clo).compress_events  = False;
