@@ -27,6 +27,7 @@
 */
 
 #include "global.h"
+#include "hash.h"
 
 /*------------------------------------------------------------*/
 /*--- Tracing Data Allocations                             ---*/
@@ -57,22 +58,21 @@ struct _SubEntry {
 #define TYPE_STACK     4
 
 struct _DataType {
+  // this is an entry of a rhash
+  rhash_entry* next;
+  UInt hash;
+
+  char* name;
   UInt number;
-  
-  ULong* counters; /* allocated lazy */
+
+  ULong read_cost, write_cost;
+  UInt allocCount, freeCount, allocSize, freeSize;
 
   UInt size;
   UInt subentries, capacity;
   SubEntry e[1];
 };
 
-typedef struct _TablePos TablePos;
-struct _TablePos {
-  UInt addr;
-  UInt pos1, pos2, pos3;
-  UInt *table2, *table3;
-};
-
 struct _DataObject {
   DataType* type;
   Char* name;
@@ -81,6 +81,123 @@ struct _DataObject {
   ULong* cost; /* allocated lazy */
 };
 
+
+/*------------------------------------------------------------*/
+/*--- Data types and objects                               ---*/
+/*------------------------------------------------------------*/
+
+UInt type_counter = 0;
+
+static UInt str_hash(const UChar *s)
+{
+    int hash_value = 0;
+    for ( ; *s; s++)
+        hash_value = (253 * hash_value + *s);
+    return hash_value;
+}
+
+Bool is_type(rhash_entry* e, void* data)
+{
+  DataObject* o = (DataObject*) data;
+  DataType* t   = (DataType*) e;
+  return (VG_(strcmp)(o->name, t->name)==0);
+}
+
+rhash_entry* new_type(void* data)
+{
+  DataObject* o = (DataObject*) data;
+  DataType* new = VG_(malloc)(sizeof(DataType));
+  new->hash   = str_hash(o->name);
+  new->name   = VG_(strdup)(o->name);
+  new->number = type_counter++;
+  new->read_cost = 0;
+  new->write_cost = 0;
+  new->allocCount = 0;
+  new->freeCount = 0;
+  new->allocSize = 0;
+  new->freeSize = 0;
+
+  CT_DEBUG(3," new_type(%s)\n", new->name);
+
+  return (rhash_entry*) new;
+}
+
+
+rhash type_hash = {
+  .name = "TypeHash",
+  .size = 577,
+  .new_entry = new_type,
+  .has_key = is_type
+};
+
+static
+void init_type_table()
+{
+  SK_(rhash_init)(&type_hash);
+}
+  
+static
+DataType* get_type(DataObject* o)
+{
+  return (DataType*) SK_(rhash_get)(&type_hash, str_hash(o->name), o);
+}
+
+static
+void print_type(rhash_entry* e)
+{
+  DataType* t = (DataType*) e;
+ 
+  CT_DEBUG(0,"T %5d /%4d Objs %6d /%5d %8llu R %8llu W %s\n",
+	   t->allocCount, t->allocCount - t->freeCount,
+	   t->allocSize, t->allocSize - t->freeSize,
+	   t->read_cost, t->write_cost, t->name);
+}
+
+static
+void print_type_table()
+{
+  CT_DEBUG(0,"Entries of Type Table (%d):\n", type_hash.entries);
+  SK_(rhash_forall)(&type_hash, print_type);
+}
+
+static
+DataObject* new_object(char* name, UInt addr, UInt size)
+{
+  DataObject* o = (DataObject*) VG_(malloc)(sizeof(DataObject));
+
+  o->name  = VG_(strdup)(name);
+  o->addr  = addr;
+  o->size  = size;
+  o->cost  = 0;
+  o->type  = get_type(o);
+
+  CT_DEBUG(3," new_object(%s): addr %p, size %d\n", o->name, o->addr, o->size);
+
+  return o;
+}
+
+
+/*------------------------------------------------------------*/
+/*--- Page Table                                           ---*/
+/*------------------------------------------------------------*/
+
+/*
+ * The page table is used for fast mapping of an address to
+ * an DataObject, which covers an address range.
+ *
+ * Functions:
+ *  void add_object(DataObject*)
+ *  void remove_object(DataObject*)
+ *  DataObject* get_object(addr)
+ */
+
+typedef struct _TablePos TablePos;
+struct _TablePos {
+  UInt addr;
+  UInt pos1, pos2, pos3;
+  UInt *table2, *table3;
+};
+
 /* 2 lower bits of each table entry encode the type.
  * The pointers itself have to be 4-byte aligned.
  */
@@ -102,53 +219,14 @@ struct _DataObject {
 
 #define TABLE_LEVEL3_SIZE 1024
 
-
-struct _TypeList {
-  int entries, capacity;
-  DataType* t[1];
-};
-
-struct _ObjectList {
-  int entries, capacity;
-  DataObject* o[1];
-};
-
-#define INITIAL_OBJECTLIST_SIZE 1024
-#define INITIAL_TYPELIST_SIZE 1024
-
-
 UInt  data_table[TABLE_LEVEL1_SIZE];
-TypeList*   types = 0;
-ObjectList* objects = 0;
 
-void init_data()
+void init_page_table()
 {
   Int i;
 
   for(i=0;i<TABLE_LEVEL1_SIZE;i++)
     data_table[i] = 0;
-
-  types = (TypeList*) VG_(malloc)(sizeof(TypeList) +
-				  INITIAL_TYPELIST_SIZE *
-				  sizeof(DataType*));
-  types->capacity = INITIAL_TYPELIST_SIZE;
-  types->entries = 0;
-  for(i=0;i<INITIAL_TYPELIST_SIZE;i++)
-    types->t[i] = 0;
-
-  objects = (ObjectList*) VG_(malloc)(sizeof(ObjectList) +
-				      INITIAL_OBJECTLIST_SIZE *
-				      sizeof(DataObject*));
-  objects->capacity = INITIAL_OBJECTLIST_SIZE;
-  objects->entries = 0;
-  for(i=0;i<INITIAL_OBJECTLIST_SIZE;i++)
-    objects->o[i] = 0;
-}
-
-static
-DataType* get_type(const char* s)
-{
-  return 0;
 }
 
 static
@@ -177,19 +255,7 @@ UInt* new_l3table()
   return t;
 }
 
-static
-DataObject* new_object(char* name, UInt addr, UInt size)
-{
-  DataObject* o = (DataObject*) VG_(malloc)(sizeof(DataObject));
 
-  o->type  = 0;
-  o->name  = VG_(strdup)(name);
-  o->addr  = addr;
-  o->size  = size;
-  o->cost  = 0;
-
-  return o;
-}
 
 static UInt tablePosAddr(UInt pos1, UInt pos2, UInt pos3)
 {
@@ -321,7 +387,7 @@ void compactTable2(UInt* table, int idx)
   if ((table[idx] & ENTRY_TYPE_MASK) != ENTRY_SUBTABLE) return;
   subtable = (UInt*)(table[idx] & ~3);
   for(i=0;i<TABLE_LEVEL3_SIZE;i++) {
-    if ((subtable[i] & ENTRY_TYPE_MASK) != ENTRY_INVALID)
+    if (subtable[i] != ENTRY_INVALID)
       isEmpty = False;
   }
   if (!isEmpty) return;
@@ -343,7 +409,7 @@ void compactTable1(UInt* table, int idx)
   subtable = (UInt*)(table[idx] & ~3);
   for(i=0;i<TABLE_LEVEL2_SIZE;i++) {
     compactTable2(subtable,i);
-    if ((subtable[i] & ENTRY_TYPE_MASK) != ENTRY_INVALID)
+    if (subtable[i] != ENTRY_INVALID)
       isEmpty = False;
   }
   if (!isEmpty) return;
@@ -355,154 +421,48 @@ void compactTable1(UInt* table, int idx)
 }
     
 
+/* fill all entries between start and end */
 static
-Bool hasOnlyEntry(UInt entry, TablePos* start, TablePos* end)
+void changeEntry(UInt from, UInt to, TablePos* start, TablePos* end)
 {
   int i;
   int start_pos1, end_pos1;
   int start_pos2, end_pos2;
 
-  CT_DEBUG(5," hasOnlyEntry(%p, %d/%d/%d - %d/%d/%d)\n",
-	   entry,
+  CT_DEBUG(5," changeEntry(%p -> %p, %d/%d/%d - %d/%d/%d)\n",
+	   from, to,
 	   start->pos1, start->pos2, start->pos3,
 	   end->pos1, end->pos2, end->pos3);
 
   if (start->pos1 == end->pos1) {
     if (start->pos2 == end->pos2) {
-      CT_DEBUG(5,"  Checking (%d/%d/%d-%d)\n",
-	       start->pos1, start->pos2, start->pos3, end->pos3);
-      for(i = start->pos3; i<=end->pos3; i++)
-	if (start->table3[i] != entry) {
-	  CT_DEBUG(5,"  => No (at %d)\n", i);
-	  return False;
-	}
+      for(i = start->pos3; i<=end->pos3; i++) {
+	CT_ASSERT(start->table3[i] == from);
+	start->table3[i] = to;
+      }
     }
     else {
       start_pos2 = start->pos2, end_pos2 = end->pos2;
       if (start->table3) {
-	CT_DEBUG(5,"  Checking (%d/%d/%d-%d)\n",
-		 start->pos1, start->pos2, start->pos3, TABLE_LEVEL3_SIZE-1);
 	start_pos2++;
-	for(i = start->pos3; i<TABLE_LEVEL3_SIZE; i++)
-	  if (start->table3[i] != entry) {
-	    CT_DEBUG(5,"  => No (at %d)\n", i);
-	    return False;
-	  }
+	for(i = start->pos3; i<TABLE_LEVEL3_SIZE; i++) {
+	  CT_ASSERT(start->table3[i] == from);
+	  start->table3[i] = to;
+	}
       }
       if (end->table3) {
-	CT_DEBUG(5,"  Checking (%d/%d/%d-%d)\n",
-		 start->pos1, end->pos2, 0, end->pos3);
 	end_pos2--;
-	for(i = 0; i<= end->pos3; i++)
-	  if (end->table3[i] != entry) {
-	    CT_DEBUG(5,"  => No (at %d)\n", i);
-	    return False;
-	  }
+	for(i = 0; i<= end->pos3; i++) {
+	  CT_ASSERT(end->table3[i] == from);
+	  end->table3[i] = to;
+	}
       }
-      if (start_pos2 <= end_pos2)
-	CT_DEBUG(5,"  Checking (%d/%d-%d)\n", start->pos1, start_pos2, end_pos2);
       for(i = start_pos2; i<=end_pos2; i++) {
-	compactTable2(start->table2,i);
-	if (start->table2[i] != entry) {
-	  CT_DEBUG(5,"  => No (at %d)\n", i);
-	  return False;
-	}
+	if (from == ENTRY_INVALID) compactTable2(start->table2,i);
+	CT_ASSERT(start->table2[i] == from);
+	start->table2[i] = to;
       }
     }
-    CT_DEBUG(5,"  => Yes\n");
-    return True;
-  }
-
-  start_pos1 = start->pos1, end_pos1 = end->pos1;
-  if (start->table2) {
-    start_pos1++;
-    start_pos2 = start->pos2;
-    if (start->table3) {
-      CT_DEBUG(5,"  Checking (%d/%d/%d-%d)\n", start->pos1, start->pos2, start->pos3,
-	       TABLE_LEVEL3_SIZE-1);
-      start_pos2++;      
-      for(i = start->pos3; i<TABLE_LEVEL3_SIZE; i++)
-	if (start->table3[i] != entry) {
-	  CT_DEBUG(5,"  => No (at %d)\n", i);
-	  return False;
-	}
-    }
-    CT_DEBUG(5,"  Checking (%d/%d-%d)\n", start->pos1, start_pos2,
-	     TABLE_LEVEL2_SIZE-1);
-    for(i = start_pos2; i<TABLE_LEVEL2_SIZE; i++) {
-      compactTable2(start->table2,i);
-      if (start->table2[i] != entry) {
-	CT_DEBUG(5,"  => No (at %d)\n", i);
-	return False;
-      }
-    }
-  }
-  if (end->table2) {
-    end_pos1++;
-    end_pos2 = end->pos2;
-    if (end->table3) {
-      CT_DEBUG(5,"  Checking (%d/%d/%d-%d)\n", end->pos1, end->pos2, 0, end->pos3);
-      end_pos2--;
-      for(i = 0; i<= end->pos3; i++)
-	if (end->table3[i] != entry) {
-	  CT_DEBUG(5,"  => No (at %d)\n", i);
-	  return False;
-	}
-    }
-    CT_DEBUG(5,"  Checking (%d/%d-%d)\n", end->pos1, 0, end_pos2);
-    for(i = 0; i<=end_pos2; i++) {
-      compactTable2(end->table2,i);
-      if (end->table2[i] != entry) {
-	CT_DEBUG(5,"  => No (at %d)\n", i);
-	return False;
-      }
-    }
-  }
-  if (start_pos1 <= end_pos1)
-    CT_DEBUG(5,"  Checking (%d-%d)\n", start_pos1, end_pos1);
-  for(i = start_pos1; i<=end_pos1; i++) {
-    compactTable1(data_table,i);
-    if (data_table[i] != entry) {
-      CT_DEBUG(5,"  => No (at %d)\n", i);
-      return False;
-    }
-  }
-  CT_DEBUG(5,"  => Yes\n");
-  return True;
-}
-
-static
-void setEntry(UInt entry, TablePos* start, TablePos* end)
-{
-  int i;
-  int start_pos1, end_pos1;
-  int start_pos2, end_pos2;
-
-  CT_DEBUG(5," setEntry(%p, %d/%d/%d - %d/%d/%d)\n",
-	   entry,
-	   start->pos1, start->pos2, start->pos3,
-	   end->pos1, end->pos2, end->pos3);
-
-  if (start->pos1 == end->pos1) {
-    if (start->pos2 == end->pos2) {
-      for(i = start->pos3; i<=end->pos3; i++)
-	start->table3[i] = entry;
-    }
-    else {
-      start_pos2 = start->pos2, end_pos2 = end->pos2;
-      if (start->table3) {
-	start_pos2++;
-	for(i = start->pos3; i<TABLE_LEVEL3_SIZE; i++)
-	  start->table3[i] = entry;
-      }
-      if (end->table3) {
-	end_pos2--;
-	for(i = 0; i<= end->pos3; i++)
-	  end->table3[i] = entry;
-      }
-      for(i = start_pos2; i<=end_pos2; i++)
-	start->table2[i] = entry;
-    }
     return;
   }
 
@@ -512,25 +472,38 @@ void setEntry(UInt entry, TablePos* start, TablePos* end)
     start_pos2 = start->pos2;
     if (start->table3) {
       start_pos2++;      
-      for(i = start->pos3; i<TABLE_LEVEL3_SIZE; i++)
-	start->table3[i] = entry;
+      for(i = start->pos3; i<TABLE_LEVEL3_SIZE; i++) {
+	CT_ASSERT(start->table3[i] == from);
+	start->table3[i] = to;
+      }
+    }
+    for(i = start_pos2; i<TABLE_LEVEL2_SIZE; i++) {
+      if (from == ENTRY_INVALID) compactTable2(start->table2,i);
+      CT_ASSERT(start->table2[i] == from);
+      start->table2[i] = to;
     }
-    for(i = start_pos2; i<TABLE_LEVEL2_SIZE; i++)
-      start->table2[i] = entry;
   }
   if (end->table2) {
-    end_pos1++;
+    end_pos1--;
     end_pos2 = end->pos2;
     if (end->table3) {
       end_pos2--;
-      for(i = 0; i<= end->pos3; i++)
-	end->table3[i] = entry;
+      for(i = 0; i<= end->pos3; i++) {
+	CT_ASSERT(end->table3[i] == from);
+	end->table3[i] = to;
+      }
+    }
+    for(i = 0; i<=end_pos2; i++) {
+      if (from == ENTRY_INVALID) compactTable2(end->table2,i);
+      CT_ASSERT(end->table2[i] == from);
+      end->table2[i] = to;
     }
-    for(i = 0; i<=end_pos2; i++)
-      end->table2[i] = entry;
   }
-  for(i = start_pos1; i<=end_pos1; i++)
-    data_table[i] = entry;
+  for(i = start_pos1; i<=end_pos1; i++) {
+    if (from == ENTRY_INVALID) compactTable1(data_table,i);
+    CT_ASSERT(data_table[i] == from);
+    data_table[i] = to;
+  }
 }
 
 /* Sets start and end according to addr and size */
@@ -570,14 +543,14 @@ Bool setRange(UInt addr, UInt size, TablePos* start, TablePos* end)
 
 
 static
-void printTable()
+void print_page_table()
 {
   int pos1, pos2, pos3;
   UInt entry1, entry2, entry3, *table2, *table3;
   DataObject* o;
   UInt last1, last2, last3;
 
-  CT_DEBUG(3," printTable:\n");
+  CT_DEBUG(0,"Page Table:\n");
 
   last1 = 0;
   for(pos1=0; pos1<TABLE_LEVEL1_SIZE; pos1++) {
@@ -586,23 +559,23 @@ void printTable()
     if ((entry1 == last1) &&
 	(entry1 & ENTRY_TYPE_MASK) != ENTRY_SUBTABLE) continue;
     if (last1 != 0)
-      CT_DEBUG(3,"  ... - (%d)\n", pos1-1);
+      CT_DEBUG(0,"P  ... - (%d)\n", pos1-1);
     last1 = entry1;
 
     if ((entry1 & ENTRY_TYPE_MASK) == ENTRY_OBJECT) {
       o = (DataObject*)(entry1 & ~3);
-      CT_DEBUG(3,"  at L1 (%d): (%p/%u) %s\n",
+      CT_DEBUG(0,"P  at L1 (%d): (%p/%u) %s\n",
 	       pos1, o->addr, o->size, o->name); 
       continue;
     }
     if ((entry1 & ENTRY_TYPE_MASK) == ENTRY_INVALID) {
-      CT_DEBUG(3,"  at L1 (%d): None\n", pos1);
+      CT_DEBUG(0,"P  at L1 (%d): None\n", pos1);
       continue;
     }
     last1 = 0;
 
     table2 = (UInt*)(entry1 & ~3);
-    CT_DEBUG(3,"  subtable (%p) at L1 %d (%p-%p)\n",
+    CT_DEBUG(0,"P  subtable (%p) at L1 %d (%p-%p)\n",
 	     table2, pos1,
 	     tablePosAddr(pos1,0,0),
 	     tablePosAddr(pos1,TABLE_LEVEL2_SIZE-1, TABLE_LEVEL3_SIZE-1));
@@ -614,23 +587,23 @@ void printTable()
       if ((entry2 == last2) &&
 	  (entry2 & ENTRY_TYPE_MASK) != ENTRY_SUBTABLE) continue;
       if (last2 != 0)
-	CT_DEBUG(3,"   ... - (%d/%d)\n", pos1, pos2-1);
+	CT_DEBUG(0,"P   ... - (%d/%d)\n", pos1, pos2-1);
       last2 = entry2;
 
       if ((entry2 & ENTRY_TYPE_MASK) == ENTRY_OBJECT) {
 	o = (DataObject*)(entry2 & ~3);
-	CT_DEBUG(3,"   at L2 (%d/%d): (%p/%u) %s\n",
+	CT_DEBUG(0,"P   at L2 (%d/%d): (%p/%u) %s\n",
 		 pos1, pos2, o->addr, o->size, o->name); 
 	continue;
       }
       if ((entry2 & ENTRY_TYPE_MASK) == ENTRY_INVALID) {
-	CT_DEBUG(3,"   at L2 (%d/%d): None\n", pos1, pos2);
+	CT_DEBUG(3,"P   at L2 (%d/%d): None\n", pos1, pos2);
 	continue;
       }
       last2 = 0;
 
       table3 = (UInt*)(entry2 & ~3);
-      CT_DEBUG(3,"   subtable (%p) at L2 %d/%d (%p-%p)\n", 
+      CT_DEBUG(0,"P   subtable (%p) at L2 %d/%d (%p-%p)\n", 
 	       table3, pos1, pos2,
 	       tablePosAddr(pos1,pos2,0),
 	       tablePosAddr(pos1,pos2, TABLE_LEVEL3_SIZE-1));
@@ -641,53 +614,21 @@ void printTable()
 
 	if (entry3 == last3) continue;
 	if (last3 != 0)
-	  CT_DEBUG(3,"    ... - (%d/%d/%d)\n", pos1, pos2, pos3-1);
+	  CT_DEBUG(0,"P    ... - (%d/%d/%d)\n", pos1, pos2, pos3-1);
 	last3 = entry3;
 
 	if ((entry3 & ENTRY_TYPE_MASK) == ENTRY_OBJECT) {
 	  o = (DataObject*)(entry3 & ~3);
-	  CT_DEBUG(3,"    at L3 (%d/%d/%d): (%p/%u) %s\n",
+	  CT_DEBUG(0,"P    at L3 (%d/%d/%d): (%p/%u) %s\n",
 		   pos1, pos2, pos3, o->addr, o->size, o->name); 
 	}
 	else
-	  CT_DEBUG(3,"    at L3 (%d/%d/%d): None\n", pos1, pos2, pos3);
+	  CT_DEBUG(0,"P    at L3 (%d/%d/%d): None\n", pos1, pos2, pos3);
       }
     }
   }
 }
 
-/* Insert an object into the data table
- */
-static
-DataObject* add_object(Char* name, UInt addr, UInt size)
-{
-  DataObject* o;
-  TablePos start, end;
-
-  CT_DEBUG(5," add_object(%s, addr %p, size %u)\n", name, addr, size);
-  if (!setRange(addr,size,&start,&end)) return 0;
-
-  /* check for free range */
-  if (!hasOnlyEntry(ENTRY_INVALID, &start, &end)) return 0;
-
-  o = new_object(name, addr, size);
-  setEntry( ((UInt)o) | ENTRY_OBJECT, &start, &end);
-
-  return o;
-}
-
-static
-void remove_object(DataObject* o)
-{
-  TablePos start, end;
-
-  CT_DEBUG(5," remove_object(%s, addr %p, size %u)\n", 
-	   o->name, o->addr, o->size);
-  if (!setRange(o->addr,o->size, &start, &end)) return;
-  if (!hasOnlyEntry(((UInt)o) | ENTRY_OBJECT, &start, &end)) return;
-  setEntry( ENTRY_INVALID, &start, &end);
-}
-
 DataObject* get_object(UInt addr)
 {
   int pos1, pos2, pos3;
@@ -735,6 +676,61 @@ DataObject* get_object(UInt addr)
   return 0;
 }
 
+static
+void remove_object(DataObject* o)
+{
+  TablePos start, end;
+  Bool noConflict;
+
+  CT_DEBUG(5," remove_object(%s, addr %p, size %u)\n", 
+	   o->name, o->addr, o->size);
+
+  noConflict = setRange(o->addr,o->size,&start,&end);
+  CT_ASSERT(noConflict == True);
+  changeEntry( ((UInt)o) | ENTRY_OBJECT, ENTRY_INVALID, &start, &end);
+}
+
+
+
+/* Insert an object into the data table
+ */
+static
+void add_object(DataObject* o)
+{
+  TablePos start, end;
+  Bool noConflict;
+
+  DataObject* o2 = get_object(o->addr);
+  if (o2) {
+    CT_DEBUG(0,"WARNING: removing old '%s' (%p/%d)\n         for new '%s' (%p/%d)\n",
+	     o2->name, o2->addr, o2->size,
+	     o->name, o->addr, o->size);
+    remove_object(o2);
+  }
+
+  CT_DEBUG(5," add_object(%s, addr %p, size %u)\n",
+	   o->name, o->addr, o->size);
+  noConflict = setRange(o->addr,o->size,&start,&end);
+  if (!noConflict) { print_page_table(); CT_ASSERT(0);  }
+  changeEntry( ENTRY_INVALID, ((UInt)o) | ENTRY_OBJECT, &start, &end);
+}
+
+
+
+/*------------------------------------------------------------*/
+/*--- Data handling                                        ---*/
+/*------------------------------------------------------------*/
+
+DataType* aType;
+
+void SK_(init_data)()
+{
+  init_page_table();
+  init_type_table();
+
+  aType = get_type(new_object("???",0,0));
+}
+
 
 /* Need to get the size and subelements first time
  * from debug info.
@@ -756,9 +752,16 @@ int allocated = 0, freed = 0;
 
 void SK_(handle_malloc)(UInt addr, UInt size)
 {
+  DataObject* o;
   fn_node** pfn = SK_(current_fn_stack).top;
   int opos = 0, d=0;
 
+  if (addr == 0) {
+    CT_DEBUG(3," Failed allocation (addr %x, size %d)\n",
+	     addr, size);
+    return;
+  }
+
   CT_DEBUG(3," Allocated (%x-%x, size %d)\n", addr, addr+size-1, size);
   while(*pfn) {
     CT_DEBUG(3,"  from %s\n", (*pfn)->name);
@@ -766,9 +769,42 @@ void SK_(handle_malloc)(UInt addr, UInt size)
       opos += VG_(sprintf)(obj_name+opos, "'%s", (*pfn)->name);
     pfn--;
   }
-  add_object(obj_name+1, addr, size);
+  o = new_object(obj_name+1, addr, size);
+  add_object(o);
 
-  if (((++allocated)%5)==0) printTable();
+  o->type->allocCount++;
+  o->type->allocSize += size;
+}
+
+Int SK_(handle_realloc)(UInt newaddr, UInt oldaddr, UInt newsize)
+{
+  DataObject* o;
+  Int oldsize = 0;
+
+  if (newaddr == 0) {
+    CT_DEBUG(3," Failed reallocation (newaddr %x, oldaddr %x, newsize %d)\n",
+	     newaddr, oldaddr, newsize);
+    return 0;
+  }
+
+  CT_DEBUG(3," Reallocated (old %x, new %x-%x, size %d)\n",
+	   oldaddr, newaddr, newaddr+newsize-1, newsize);
+
+  o = get_object(oldaddr);
+  if (o && (o->addr == oldaddr)) {
+    CT_DEBUG(3,"  old size %d, obj %s\n", o->size, o->name);
+    oldsize = o->size;
+    remove_object(o);
+    o->addr = newaddr;
+    o->size = newsize;
+    add_object(o);
+
+    o->type->allocSize += newsize-oldsize;
+  }
+  else
+    CT_DEBUG(3,"  nothing found to realloc.\n");
+
+  return oldsize;
 }
 
 /* returns size of freed allocation */
@@ -788,9 +824,12 @@ Int SK_(handle_free)(UInt addr)
   if (o && (o->addr == addr)) {
     remove_object(o);
     res = o->size;
-    VG_(free)(o);
 
-    if (((++freed)%5)==0) printTable();
+    o->type->freeCount++;
+    o->type->freeSize += o->size;
+
+    VG_(free)(o->name);
+    VG_(free)(o);
   }
   else
     CT_DEBUG(3,"  nothing found to free.\n");
@@ -802,3 +841,118 @@ ULong* getcost(Addr addr, Context* cxt)
 {
   return 0;
 }
+
+void SK_(handle_read)(UInt addr, UInt size, int sim)
+{
+  DataObject* o = get_object(addr);
+  DataType* t = o ? o->type : aType;
+
+  t->read_cost++;
+
+  if (o) {
+    CT_DEBUG(3,"Read  %p: Off %d in %s (%p/%d)\n",
+	     addr, addr - o->addr, o->name, o->addr, o->size);
+  }
+  CT_DEBUG(3,"Read (%p), reads of %s: %llu\n", addr, t->name, t->read_cost);
+}
+
+void SK_(handle_write)(UInt addr, UInt size, int sim)
+{
+  DataObject* o = get_object(addr);
+  DataType* t = o ? o->type : aType;
+
+  t->write_cost++;
+
+  if (o) {
+    CT_DEBUG(3,"Write %p: Off %d in %s (%p/%d)\n",
+	     addr, addr - o->addr, o->name, o->addr, o->size);
+  }
+  CT_DEBUG(3,"Write (%p), writes of %s: %llu\n", addr, t->name, t->write_cost);
+}
+
+void SK_(finish_data)()
+{
+  print_type_table();
+  //print_page_table();
+}
+
+void SK_(my_new_mem_startup)(Addr a, SizeT len, Bool rr, Bool ww, Bool xx)
+{
+  //DataObject* o;
+  static char buf[64];
+
+  if (!SK_(clo).collect_data) return;
+
+  VG_(sprintf)(buf,"Seg[%08p-%08p/%c%c%c]", a, a+len-1,
+	       rr ? 'r':'-', ww ? 'w':'-', xx ? 'x':'-');
+
+  CT_DEBUG(0,"new_mem_startup: %s\n", buf);
+  
+  //o = new_object(buf, a, len);
+  //add_object(o);
+}
+
+void SK_(my_new_mem_stack_signal)(Addr a, SizeT len)
+{
+  DataObject* o;
+  static char buf[64];
+
+  if (!SK_(clo).collect_data) return;
+
+  VG_(sprintf)(buf,"Stack[%08p-%08p]", a, a+len-1);
+
+  CT_DEBUG(0,"new_mem_stack_signal: %s\n", buf);
+  
+  o = new_object(buf, a, len);
+  add_object(o);
+}
+
+void SK_(my_new_mem_brk)(Addr a, SizeT len)
+{
+  DataObject* o;
+  static char buf[64];
+
+  if (!SK_(clo).collect_data) return;
+
+  VG_(sprintf)(buf,"Brk[%08p-%08p]", a, a+len-1);
+
+  CT_DEBUG(0,"new_mem_brk: %s\n", buf);
+  
+  o = new_object(buf, a, len);
+  add_object(o);
+}
+
+void SK_(my_new_mem_mmap)(Addr a, SizeT len, Bool rr, Bool ww, Bool xx)
+{
+  //DataObject* o;
+  static char buf[64];
+
+  if (!SK_(clo).collect_data) return;
+
+  VG_(sprintf)(buf,"MMap[%08p-%08p/%c%c%c]", a, a+len-1,
+	       rr ? 'r':'-', ww ? 'w':'-', xx ? 'x':'-');
+
+  CT_DEBUG(0,"new_mem_mmap: %s\n", buf);
+  
+  //o = new_object(buf, a, len);
+  //add_object(o);
+}
+
+void SK_(new_data_sym)(Addr a, SizeT len, SegInfo* si, Char* name)
+{
+  DataObject* o;
+  static char buf[256];
+
+  VG_(sprintf)(buf,"%s:%s [%08p/%d]", 
+	       VG_(seg_filename)(si), name, a, len);
+
+  CT_DEBUG(0,"new_data_sym: %s\n", buf);
+
+  o = new_object(buf, a, len);
+  add_object(o);
+}
+
+void SK_(free_seg_info)(SegInfo* si)
+{
+  CT_DEBUG(0,"free_seg_info: %s\n", VG_(seg_filename)(si));
+}
