@@ -29,6 +29,7 @@
 #include "global.h"
 #include "events.h"
 
+
 /*------------------------------------------------------------*/
 /*--- Support for signal handlers and multi-threading      ---*/
 /*------------------------------------------------------------*/
@@ -61,6 +62,11 @@ Int SK_(get_dump_counter)(void)
   return out_counter;
 }
 
+Char* SK_(get_dump_file_base)()
+{
+  return dump_file_base;
+}
+
 /*------------------------------------------------------------*/
 /*--- Output file related stuff                            ---*/
 /*------------------------------------------------------------*/
@@ -1214,7 +1220,7 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
 
     CT_ASSERT(filename != 0);
 
-    if (SK_(clo).separate_dumps) {
+    if (!SK_(clo).combine_dumps) {
 	i = VG_(sprintf)(filename, "%s.%d", dump_file_base, VG_(getpid)());
     
 	if (trigger)
@@ -1278,6 +1284,9 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
 
     /* "desc:" lines */
     if (!appending) {
+	fwrite(fd, "\n", 1);
+
+#if 0
 	/* Global options changing the tracing behaviour */
 	VG_(sprintf)(buf, "\ndesc: Option: --skip-plt=%s\n",
 		     SK_(clo).skip_plt ? "yes" : "no");
@@ -1292,7 +1301,6 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
 		     SK_(clo).separate_callers);
 	fwrite(fd, (void*)buf, VG_(strlen)(buf));
 
-#if 0
 	VG_(sprintf)(buf, "desc: Option: --dump-bbs=%s\n",
 		     SK_(clo).dump_bbs ? "yes" : "no");
 	fwrite(fd, (void*)buf, VG_(strlen)(buf));
@@ -1390,7 +1398,7 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
    fwrite(fd, "\n\n",2);
 
    if (VG_(clo_verbosity) > 1)
-       VG_(message)(Vg_DebugMsg, "  Dumping to %s...", filename);
+       VG_(message)(Vg_DebugMsg, "Dump to %s", filename);
 
    return fd;
 }
@@ -1416,9 +1424,6 @@ static void close_dumpfile(Char buf[BUF_LEN], int fd, int tid)
 			 filename, filename);
        }
    }
-
-    if (VG_(clo_verbosity) > 1)
-	VG_(message)(Vg_DebugMsg, "  Finished dump.");
 }
 
 
@@ -1537,7 +1542,7 @@ void SK_(dump_profile)(Char* trigger, Bool only_current_thread)
 	    trigger ? trigger : (Char*)"Prg.Term.");
 
    if (VG_(clo_verbosity) > 1)
-       VG_(message)(Vg_DebugMsg, "Prepare dump at BB %llu (%s)...",
+       VG_(message)(Vg_DebugMsg, "Start dumping at BB %llu (%s)...",
 		    SK_(stat).bb_executions,
 		    trigger ? trigger : (Char*)"Prg.Term.");
 
@@ -1548,7 +1553,7 @@ void SK_(dump_profile)(Char* trigger, Bool only_current_thread)
    last_bbs_done = VG_(bbs_done);
 
    if (VG_(clo_verbosity) > 1)
-     VG_(message)(Vg_DebugMsg, "... finished dumping ");
+     VG_(message)(Vg_DebugMsg, "Dumping done.");
 
    VGP_POPCC(VgpCacheDump);
 }
