@@ -16,7 +16,7 @@
 /*------------------------------------------------------------*/
 
 /* Enable debug output */
-#define CT_ENABLE_DEBUG 0
+#define CT_ENABLE_DEBUG 1
 
 /* Enable experimental features? */
 #define CT_EXPERIMENTAL 0
@@ -47,7 +47,7 @@ struct _CommandLineOptions {
 
   /* Dump format options */
   Char* filename_base;   /* Base name for dumps */
-  Bool separate_dumps;   /* Dump trace parts as separate files? */
+  Bool combine_dumps;    /* Dump trace parts into same file? */
   Bool compress_strings;
   Bool compress_events;
   Bool compress_pos;
@@ -60,6 +60,7 @@ struct _CommandLineOptions {
   
   /* Dump generation options */
   Int dump_every_bb;     /* Dump every xxx BBs. */
+  Int dump_objs;         /* Dump static object info */
   
   /* Collection options */
   Bool separate_threads; /* Separate threads in dump? */
@@ -349,6 +350,7 @@ struct _fn_node {
   Bool skip :1;
 
   Bool is_malloc :1;
+  Bool is_realloc :1;
   Bool is_free :1;
   Int  is_constructor :1;
   Int  is_destructor :1;
@@ -579,6 +581,7 @@ struct cachesim_if
   void (*printstat)(void);  
   void (*add_icost)(SimCost, BBCC*, InstrInfo*);
   void (*after_bbsetup)(void);
+  void (*finish)(void);
   
   CT_REGPARM(1) void (*log_0D)(InstrInfo* ii);
   CT_REGPARM(2) void (*log_1Dr)(InstrInfo* ii, Addr data);
@@ -591,15 +594,15 @@ struct cachesim_if
 /*--- Functions                                            ---*/
 /*------------------------------------------------------------*/
 
-/* from ct_clo.c */
+/* from clo.c */
 
 void SK_(set_clo_defaults)();
 void SK_(update_fn_config)(fn_node*);
 
 
-/* from ct_sim.c */
+/* from sim.c */
 struct event_sets {
-  EventSet *Ir, *Dr, *Dw;
+  EventSet *use, *Ir, *Dr, *Dw;
   EventSet *D0, *D1r, *D1w, *D2;
   EventSet *sim;
   EventSet *full; /* sim plus user events */
@@ -615,7 +618,7 @@ extern struct cachesim_if SK_(cachesim);
 
 void SK_(init_eventsets)(Int user);
 
-/* from ct_main.c */
+/* from main.c */
 Bool SK_(get_debug_info)(Addr, Char filename[FILENAME_LEN],
 			 Char fn_name[FN_NAME_LEN], Int*, SegInfo**);
 void SK_(set_instrument_state)(Char*,Bool);
@@ -623,12 +626,12 @@ void SK_(dump_profile)(Char* trigger,Bool only_current_thread);
 void SK_(zero_all_cost)(Bool only_current_thread);
 Int SK_(get_dump_counter)(void);
 
-/* from ct_command.c */
+/* from command.c */
 void SK_(init_command)(Char* dir, Char* dumps);
 void SK_(check_command)(void);
 void SK_(finish_command)(void);
 
-/* from ct_bb.c */
+/* from bb.c */
 void SK_(init_bb_hash)();
 bb_hash* SK_(get_bb_hash)();
 BB*  SK_(get_bb)(Addr addr, UCodeBlock* cb_in, Bool *seen_before);
@@ -638,7 +641,7 @@ static __inline__ Addr bb_addr(BB* bb)
 static __inline__ Addr bb_jmpaddr(BB* bb)
  { return bb->jmp_offset + bb->offset + bb->obj->offset; }
 
-/* from ct_fn.c */
+/* from fn.c */
 void SK_(init_fn_array)(fn_array*);
 void SK_(copy_current_fn_array)(fn_array* dst);
 fn_array* SK_(get_current_fn_array)();
@@ -650,7 +653,7 @@ obj_node* SK_(get_obj_node)(SegInfo* si);
 file_node* SK_(get_file_node)(obj_node*, Char* filename);
 fn_node*  SK_(get_fn_node)(BB* bb);
 
-/* from ct_bbcc.c */
+/* from bbcc.c */
 void SK_(init_bbcc_hash)(bbcc_hash* bbccs);
 void SK_(copy_current_bbcc_hash)(bbcc_hash* dst);
 bbcc_hash* SK_(get_current_bbcc_hash)();
@@ -662,14 +665,14 @@ BBCC* SK_(clone_bbcc)(BBCC* orig, Context* cxt, Int rec_index);
 __attribute__ ((regparm (1))) void SK_(setup_bbcc)(BB* bb);
 
 
-/* from ct_jumps.c */
+/* from jumps.c */
 void SK_(init_jcc_hash)(jcc_hash*);
 void SK_(copy_current_jcc_hash)(jcc_hash* dst);
 jcc_hash* SK_(get_current_jcc_hash)();
 void SK_(set_current_jcc_hash)(jcc_hash*);
 jCC* SK_(get_jcc)(BBCC* from, BBCC* to);
 
-/* from ct_callstack.c */
+/* from callstack.c */
 void SK_(init_call_stack)(call_stack*);
 void SK_(copy_current_call_stack)(call_stack* dst);
 void SK_(set_current_call_stack)(call_stack*);
@@ -679,7 +682,7 @@ void SK_(push_call_stack)(BBCC* from, BBCC* to, Addr esp, Bool skip);
 void SK_(pop_call_stack)();
 Int SK_(unwind_call_stack)(Addr esp);
 
-/* from ct_context.c */
+/* from context.c */
 void SK_(init_fn_stack)(fn_stack*);
 void SK_(copy_current_fn_stack)(fn_stack*);
 fn_stack* SK_(get_current_fn_stack)();
@@ -690,7 +693,7 @@ cxt_hash* SK_(get_cxt_hash)();
 Context* SK_(get_cxt)(fn_node** fn);
 void SK_(push_cxt)(fn_node* fn);
 
-/* from ct_threads.c */
+/* from threads.c */
 void SK_(init_threads)();
 thread_info** SK_(get_threads)();
 thread_info* SK_(get_current_thread)();
@@ -706,16 +709,35 @@ void SK_(pre_signal)(ThreadId tid, Int sigNum, Bool alt_stack);
 void SK_(post_signal)(ThreadId tid, Int sigNum);
 
 
-/* from ct_dump.c */
+/* from dump.c */
 extern FullCost SK_(total_cost);
 void SK_(init_files)(Char** dir, Char** file);
+Char* SK_(get_dump_file_base)();
+
+
+/* from static.c */
+void SK_(dump_obj)(obj_node*);
 
 
 /* from data.c */
+void SK_(init_data)();
+void SK_(finish_data)();
 void SK_(handle_malloc)(UInt start, UInt size);
+Int  SK_(handle_realloc)(UInt newstart, UInt oldstart, UInt size);
 Int  SK_(handle_free)(UInt addr);
 void SK_(handle_constructor)(UInt addr, Char* fullname);
 void SK_(handle_destructor)(UInt addr, Char* fullname);
+void SK_(handle_read)(UInt addr, UInt size, int sim);
+void SK_(handle_write)(UInt addr, UInt size, int sim);
+
+#if VG_CORE_INTERFACE_MAJOR_VERSION < 7
+typedef UInt SizeT;
+#endif
+
+void SK_(my_new_mem_startup)(Addr a, SizeT len, Bool rr, Bool ww, Bool xx);
+void SK_(my_new_mem_stack_signal)(Addr a, SizeT len);
+void SK_(my_new_mem_brk)(Addr a, SizeT len);
+void SK_(my_new_mem_mmap)(Addr a, SizeT len, Bool rr, Bool ww, Bool xx);
 
 /*------------------------------------------------------------*/
 /*--- Exported global variables                            ---*/
@@ -769,6 +791,7 @@ void SK_(print_context)(void);
 void SK_(print_jcc)(int s, jCC* jcc);
 void SK_(print_bbcc)(int s, BBCC* bbcc, Bool);
 void SK_(print_bbcc_fn)(BBCC* bbcc);
+void SK_(print_eventset)(int s, EventSet* es);
 void SK_(print_cost)(int s, EventSet*, ULong* cost);
 void SK_(print_bb)(int s, BB* bb);
 void SK_(print_cxt)(int s, Context* cxt, int rec_index);
