@@ -8,7 +8,7 @@
    This file is part of Callgrind, a Valgrind skin for call graph
    profiling programs.
 
-   Copyright (C) 2002-2004, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
+   Copyright (C) 2002-2005, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
 
    This skin is derived from and contains lot of code from Cachegrind
    Copyright (C) 2002 Nicholas Nethercote (njn25@cam.ac.uk)
@@ -605,7 +605,11 @@ void SK_(set_instrument_state)(Char* reason, Bool state)
 #if VG_CORE_INTERFACE_MAJOR_VERSION < 3
 Bool SK_(handle_client_request)(ThreadState* tst, UInt *args, UInt *ret)
 #else
+#if VG_CORE_INTERFACE_MAJOR_VERSION < 7
 Bool SK_(handle_client_request)(ThreadId tid, UInt *args, UInt *ret)
+#else
+Bool SK_(handle_client_request)(ThreadId tid, UWord *args, UWord *ret)
+#endif
 #endif
 {
    if (!VG_IS_SKIN_USERREQ('C','T',args[0]))
@@ -717,7 +721,9 @@ void finish()
 {
   char buf[RESULTS_BUF_LEN];
 
-  CT_DEBUG(0, "CallStack finalization:\n");
+  CT_DEBUG(0, "finish()\n");
+
+  (*SK_(cachesim).finish)();
 
   /* pop all remaining items from CallStack for correct sum
    */
@@ -809,13 +815,14 @@ void finish()
   VG_(message)(Vg_UserMsg, "");
 
   //  if (SK_(clo).simulate_cache)
-    (*SK_(cachesim).printstat)();
+  (*SK_(cachesim).printstat)();
 }
 
 
 void SK_(fini)(Int exitcode)
 {
   finish();
+  if (SK_(clo).collect_data) SK_(finish_data)();
 }
 
 
@@ -828,7 +835,7 @@ void SK_(pre_clo_init)(void)
     VG_(details_name)            ("Callgrind");
     VG_(details_version)         (VERSION);
     VG_(details_description)     ("a call-graph generating cache profiler");
-    VG_(details_copyright_author)("Copyright (C) 2002-2004, and GNU GPL'd, "
+    VG_(details_copyright_author)("Copyright (C) 2002-2005, and GNU GPL'd, "
 				  "by J.Weidendorfer, N.Nethercote et al.");
     VG_(details_bug_reports_to)  ("Josef.Weidendorfer@gmx.de");
     VG_(details_avg_translation_sizeB) ( 155 );
@@ -851,6 +858,12 @@ void SK_(pre_clo_init)(void)
     VG_(track_post_syscall)( & SK_(post_syscalltime) );
 #endif
 
+    /* for data collection */
+    VG_(init_new_mem_startup)(SK_(my_new_mem_startup));
+    VG_(init_new_mem_stack_signal)(SK_(my_new_mem_stack_signal));
+    VG_(init_new_mem_brk)(SK_(my_new_mem_brk));
+    VG_(init_new_mem_mmap)(SK_(my_new_mem_mmap));
+
     SK_(set_clo_defaults)();
 }
 
@@ -860,12 +873,10 @@ void SK_(pre_clo_init)(void)
  * But Callgrinds instrumentation does exactly this, so we want to
  * always switch off pointer checks.
  */
-#if (VG_CORE_INTERFACE_MAJOR_VERSION >5) || \
-    ((VG_CORE_INTERFACE_MAJOR_VERSION > 4) && defined(VG_UNSTABLE))
+#if (VG_CORE_INTERFACE_MAJOR_VERSION >5)
 #define NEED_POINTERCHECKOFF 1
 #endif
 
-
 #ifdef NEED_POINTERCHECKOFF
 extern Bool VG_(clo_pointercheck);
 #endif
@@ -892,16 +903,24 @@ void SK_(post_clo_init)(void)
 
    (*SK_(cachesim).post_clo_init)();
 
+#if VG_CORE_INTERFACE_MAJOR_VERSION < 7
+   // with VG 2.3, no need to register helpers any more
    VG_(register_compact_helper)((Addr) SK_(cachesim).log_0D);
    VG_(register_compact_helper)((Addr) SK_(cachesim).log_1Dr);
    VG_(register_compact_helper)((Addr) SK_(cachesim).log_1Dw);
    VG_(register_compact_helper)((Addr) SK_(cachesim).log_2D);
    VG_(register_compact_helper)((Addr) & SK_(setup_bbcc));
+#endif
 
    SK_(init_eventsets)(0);
    SK_(init_statistics)(& SK_(stat));
    SK_(init_cost_lz)( SK_(sets).full, &SK_(total_cost) );
 
+   if (SK_(clo).collect_data) {
+     VG_(needs_data_syms)();
+     SK_(init_data)();
+   }
+
    /* initialize hash tables */
    SK_(init_obj_table)();
    SK_(init_cxt_table)();
